BUILTIN_LANGUAGES = sorted(
    [
        "bash",
        "css",
        "go",
        "html",
        "java",
        "javascript",
        "json",
        "kotlin",
        "markdown",
        "python",
        "rust",
        "regex",
        "sql",
        "toml",
        "yaml",
    ]
)
